theory Final_Nontermination_Checker
  imports 
    "~~/src/HOL/Library/Code_Char"
    Nontermination_Checker_3
    "$AFP/XML/Xmlt"
begin
  
definition parse_var :: "string xmlt" where "parse_var x = (Xmlt.text ''var'' x)"
definition parse_sym :: "string xmlt" where "parse_sym x = (Xmlt.text ''sym'' x)"

partial_function (sum_bot)
  parse_term :: "(string, string) term xmlt"
where
  [code]: "parse_term x = Xmlt.options [
    (''var'', Xmlt.change parse_var Var),
    (''funApp'', Xmlt.many1 ''funApp'' parse_sym parse_term Fun)
  ] x"

partial_function (sum_bot) parse_ctxt :: "(string, string) ctxt xmlt" where
  [code]: "parse_ctxt x = Xmlt.options [
     (''box'', Xmlt.leaf ''box'' Hole),
     (''funCtxt'', Xmlt.tuple4 ''funCtxt''
       parse_sym
       (Xmlt.many ''before'' parse_term id)
       parse_ctxt
       (Xmlt.many ''after'' parse_term id)
       More)
   ] x"
  
definition parse_rule :: "(string,string) rule xmlt" where
  "parse_rule = Xmlt.pair ''rule'' parse_term parse_term Pair" 
 
definition parse_subst :: "(string,string) subst_impl xmlt" where
  "parse_subst = Xmlt.many ''subst''
     (Xmlt.pair ''substEntry'' parse_var parse_term Pair) id"
    
definition parse_step :: "((string,string)rstep_wit \<times> (string,string)term) xmlt" where
  "parse_step = Xmlt.triple ''rewriteStep'' 
      parse_rule parse_ctxt parse_term
     (\<lambda> lr C t. ((lr,C),t))" 

definition parse_loop_proof :: "(string,string) loop_proof xmlt" where
  "parse_loop_proof = Xmlt.tuple4 ''loop''
     parse_term (Xmlt.many ''derivation'' parse_step id) parse_ctxt parse_subst
     Loop_Proof"
  
partial_function (sum_bot) parse_non_term_proof :: "(string,string)non_term_proof xmlt" where
  [code]: "parse_non_term_proof x = Xmlt.options [
    (''loop'', Xmlt.change parse_loop_proof Loop),
    (''ruleRemoval'', Xmlt.pair ''ruleRemoval'' 
         (Xmlt.many ''delete'' parse_rule id)
         parse_non_term_proof
         Rule_Removal)
    ] x"
  
definition parse_cert_problem :: "((string,string)trs \<times> (string,string)non_term_proof) xmlt" where
  "parse_cert_problem = Xmlt.pair ''certificationProblem'' 
      (Xmlt.many ''input'' parse_rule id)
      parse_non_term_proof
      (\<lambda> rules prf. (set rules, prf))" 
    
definition string_to_cert_problem 
  :: "string \<Rightarrow> string + ((string,string)trs \<times> (string,string)non_term_proof)" 
  where 
  "string_to_cert_problem s = (
    case doc_of_string s of
      Inl e \<Rightarrow> Inl (''error while parsing string to xml\<newline>'' @ e)
    | Inr doc \<Rightarrow> case doc of
        XMLDOC _ xml \<Rightarrow> case parse_cert_problem xml of
          Left e \<Rightarrow> Inl (''error while parsing xml\<newline>'' @ e)
        | Right cp \<Rightarrow> Inr cp)" 
  
definition non_term_checker :: "string \<Rightarrow> string" where
  "non_term_checker s = (case string_to_cert_problem s of
     Inl e \<Rightarrow> ''error while parsing\<newline>'' @ e
   | Inr (R,prf) \<Rightarrow> main_checker R prf)" 
  
export_code non_term_checker
  in Haskell module_name Nontermination_Checker file "./Code"
  
end
