theory Loop_Checker_1
imports 
  Loops
  Matching
begin


text \<open>Witness/certificate for rewrite step: provide rule and context.\<close>
type_synonym ('f, 'v) rstep_wit = "('f, 'v) rule \<times> ('f, 'v) ctxt"

fun get_redex :: "('f,'v)ctxt \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term" where
  "get_redex (More f bef C aft) (Fun g ts) = (get_redex C (ts ! (length bef)))" 
| "get_redex C t = t"

context
  fixes R :: "('f, 'v) trs"
begin

text \<open>Check single rewrite step\<close>
fun check_rstep_wit :: "('f, 'v) rstep_wit \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "check_rstep_wit ((l, r), C) s t \<longleftrightarrow>
      (case match l (get_redex C s) of 
        None \<Rightarrow> False
      | Some \<sigma> \<Rightarrow>
         s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> (l, r) \<in> R)"

lemma check_rstep_wit:
  assumes "check_rstep_wit wit s t"
  shows "(s, t) \<in> rstep R"
proof -
  -- \<open>case analysis with only 1 case via "obtain" and "cases"\<close>
  obtain l r C where wit: "wit = ((l, r), C)" by (cases wit, auto)  
  from assms wit -- \<open>@{thm assms} refers to assumptions of theorem\<close>
  obtain \<sigma> where "s = C \<langle>l \<cdot> \<sigma>\<rangle>" and "t = C \<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R"
    by (cases "match l (get_redex C s)") (auto simp: Let_def)
  with rstep [of l r R C \<sigma>] show ?thesis by auto
qed

text \<open>Exercise: Check derivation.
  \<bullet> Define a function @{term check_derivation} of type
    @{typ "('f, 'v) term \<Rightarrow> (('f, 'v) rstep_wit \<times> ('f, 'v) term) list \<Rightarrow> ('f, 'v) term option"}
    which takes an initial term and a sequence of steps to new terms and either returns
    @{term "Some t"} if all rewrite steps have been validated and @{term t} is the last term
    in the derivation, or it returns @{const None} if some rewrite step could not be validated.
  \<bullet> Prove soundness of the function: if @{term "check_derivation s seq = Some t"}, then
     @{term s} rewrites to @{term t} in @{term "length seq"} steps.
    \<bullet> Use @{command find_consts} or @{command find_theorems} to find a constant which
      can express n-steps.
    \<bullet> Think about which induction rule to use.
    \<bullet> @{method auto} and @{method simp} will perform automatic case analysis on if-then-else
      by adjusting the splitter, e.g., "auto split: if_splits"


 \<close>
fun check_derivation
  :: "('f, 'v) term \<Rightarrow> (('f, 'v) rstep_wit \<times> ('f, 'v) term) list \<Rightarrow> ('f, 'v) term option"
  where
    "check_derivation s [] = Some s" |
    "check_derivation s ((wit, t) # seq) =
      (if check_rstep_wit wit s t then check_derivation t seq
      else None)"

find_consts "?'a \<Rightarrow> nat \<Rightarrow> ?'a"
find_theorems "_^*" "_ :: nat \<Rightarrow> _"

lemma check_derivation:
  assumes "check_derivation s seq = Some t"
  shows "(s, t) \<in> rstep R ^^ length seq"
  using assms
proof (induct s seq rule: check_derivation.induct)
  case (2 s wit u seq)
  from 2 [simplified] have wit: "check_rstep_wit wit s u"
    and result: "check_derivation u seq = Some t"
    by (auto split: if_splits)
  from check_rstep_wit [OF wit]
  have "(s, u) \<in> rstep R" .
  moreover have "(u, t) \<in> rstep R ^^ length seq"
    using 2 and wit and result by auto
  ultimately show ?case
    using relpow_Suc_I2 by fastforce
qed auto

end

datatype ('f, 'v) loop_proof = Loop_Proof
  "('f, 'v) term"
  "(('f, 'v) rstep_wit \<times> ('f, 'v) term) list"
  "('f, 'v) ctxt"
  "('v \<times> ('f, 'v)term) list"

fun check_loop :: "('f, 'v) trs \<Rightarrow> ('f, 'v) loop_proof \<Rightarrow> bool"
  where
    "check_loop R (Loop_Proof s seq C \<sigma>i) = (
      let \<sigma> = som (map_of \<sigma>i) in     
      seq \<noteq> [] \<and>
      (case check_derivation R s seq of
        None \<Rightarrow> False
      | Some t \<Rightarrow> t = C\<langle>s \<cdot> \<sigma>\<rangle>))"

lemma check_loop:
  assumes "check_loop R prf"
  shows "\<exists> t C \<sigma>. (t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
proof (cases "prf")
  case (Loop_Proof s seq C \<sigma>i)
  note ok = assms [unfolded Loop_Proof]
  from ok have len: "length seq \<noteq> 0" by auto
  obtain t where check: "check_derivation R s seq = Some t"
    using ok by (auto split: option.splits)
  from check_derivation [OF check]
  have "(s, t) \<in> rstep R ^^ length seq" .
  with len have steps: "(s, t) \<in> (rstep R)^+"
    by (metis relpow_E trancl_power zero_less_Suc)
  from ok and check obtain C \<sigma> where t: "t = C\<langle>s \<cdot> \<sigma>\<rangle>" by auto
  from steps t show ?thesis by blast
qed

lemma check_loop_not_SN:
  assumes "check_loop R prf"
  shows "\<not> SN (rstep R)"
  using check_loop [OF assms] and loop_imp_not_SN_rstep by blast

end