theory Loop_Checker_2
  imports
    Loops
    Matching
    "$AFP/Show/Show_Instances"
    "$AFP/Certification_Monads/Check_Monad"
begin

text \<open>Witness/certificate for rewrite step: provide rule and context.\<close>
type_synonym ('f, 'v) rstep_wit = "('f, 'v) rule \<times> ('f, 'v) ctxt"

fun get_redex :: "('f,'v)ctxt \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term" where
  "get_redex (More f bef C aft) (Fun g ts) = (get_redex C (ts ! (length bef)))" 
| "get_redex C t = t"

context fixes
  R :: "('f,'v)trs"
begin

text \<open>Check single rewrite step\<close>
fun check_rstep_wit :: "('f,'v)rstep_wit \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> shows check" where
  "check_rstep_wit ((l,r), C) s t = (
      case match l (get_redex C s) of 
        None \<Rightarrow> error (shows ''problem with matching'')
      |  Some \<sigma> \<Rightarrow> do {
      let s' = C \<langle>l \<cdot> \<sigma>\<rangle>;
      let t' = C \<langle>r \<cdot> \<sigma>\<rangle>;
      check (s = s') (shows ''problem with s'');
      check (t = t') (shows ''problem with t'');
      check ((l,r) \<in> R) (shows ''(l,r) is not a rewrite rule of TRS'')
    }) <+? (\<lambda> e. shows ''error in checking rewrite step s -> t\<newline>'' o e)"

lemma check_rstep_wit: assumes "isOK(check_rstep_wit wit s t)"
  shows "(s,t) \<in> rstep R"
proof -
  -- \<open>case analysis with only 1 case via "obtain" and "cases"\<close>
  obtain l r C where wit: "wit = ((l, r), C)" by (cases wit, auto)  
  from assms wit -- \<open>@{thm assms} refers to assumptions of theorem\<close>
  obtain \<sigma> where "s = C \<langle>l \<cdot> \<sigma>\<rangle>" and "t = C \<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R"
    by (cases "match l (get_redex C s)") (auto simp: Let_def)
  with rstep [of l r R C \<sigma>] show ?thesis by auto
qed

fun check_derivation
  :: "('f,'v)term \<Rightarrow> (('f,'v)rstep_wit \<times> ('f,'v)term)list \<Rightarrow>  shows + ('f,'v)term"
where
  "check_derivation s [] = return s" |
  "check_derivation s ((wit,t) # seq) = (do {
    check_rstep_wit wit s t;
    check_derivation t seq})"

lemma check_derivation: assumes "check_derivation s seq = return t"
  shows "(s,t) \<in> (rstep R)^^(length seq)"
  using assms
proof (induct s seq rule: check_derivation.induct)
  case (2 s wit u seq)
  from 2[simplified] have wit: "isOK(check_rstep_wit wit s u)"
    and result: "check_derivation u seq = return t"
    by (auto split: if_splits)
  from check_rstep_wit[OF wit]
  have "(s, u) \<in> rstep R" .
  moreover have "(u,t) \<in> (rstep R)^^(length seq)"
    using 2 wit result by auto
  ultimately show ?case
    using relpow_Suc_I2 by fastforce
qed auto
end

type_synonym ('f,'v)subst_impl = "('v \<times> ('f, 'v) term) list" 
  
datatype ('f, 'v) loop_proof = Loop_Proof
  "('f, 'v) term"
  "(('f, 'v) rstep_wit \<times> ('f, 'v) term) list"
  "('f, 'v) ctxt"
  "('f,'v) subst_impl"

fun check_loop :: "('f,'v)trs \<Rightarrow> ('f,'v)loop_proof \<Rightarrow> shows check" where
  "check_loop R (Loop_Proof s seq C \<sigma>i) = do {
      let \<sigma> = som (map_of \<sigma>i);
      check (seq \<noteq> []) (shows ''loop must not be empty'');
      t \<leftarrow> check_derivation R s seq
         <+? (\<lambda> e. shows ''error in checking derivation\<newline>'' o e);
      check (t = C \<langle> s \<cdot> \<sigma>\<rangle>)
        (shows ''last term t does not have expected form'')}"

lemma check_loop:
  assumes "isOK(check_loop R prf)"
  shows "\<exists> t C \<sigma>. (t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
proof (cases "prf")
  case (Loop_Proof s seq C \<sigma>i)
  note ok = assms [unfolded Loop_Proof]
  from ok have len: "length seq \<noteq> 0" by auto
  obtain t where check: "check_derivation R s seq = return t"
    using ok by (auto split: option.splits)
  from check_derivation [OF check]
  have "(s, t) \<in> rstep R ^^ length seq" .
  with len have steps: "(s, t) \<in> (rstep R)^+"
    by (metis relpow_E trancl_power zero_less_Suc)
  from ok and check obtain C \<sigma> where t: "t = C\<langle>s \<cdot> \<sigma>\<rangle>" by auto
  from steps t show ?thesis by blast
qed

lemma check_loop_not_SN: assumes "isOK(check_loop R prf)"
  shows "\<not> SN (rstep R)"
  using check_loop[OF assms] loop_imp_not_SN_rstep by blast

end