theory Loop_Checker_Sound
imports 
  Loop_Checker_Demo
  Loops
begin
  
section \<open>Soundness proof\<close>

lemma aim: "check_loop R prf \<Longrightarrow> \<not> SN (rstep R)" 
  oops
    
lemma aim_1: "check_loop R prf \<Longrightarrow> \<exists> t C \<sigma>. (t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
  oops
    
lemma aim_2: "(t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+ \<Longrightarrow> \<not> SN (rstep R)" 
  oops

text \<open>Start with checking rewrite step\<close>
lemma check_rstep_wit:
  assumes "check_rstep_wit R wit s t"
  shows "(s, t) \<in> rstep R"
proof -
  -- \<open>case analysis with only 1 case via "obtain" and "cases"\<close>
  obtain l r C where wit: "wit = ((l, r), C)" by (cases wit, auto)  
  from assms wit -- \<open>@{thm assms} refers to assumptions of theorem\<close>
  obtain \<sigma> where "s = C \<langle>l \<cdot> \<sigma>\<rangle>" and "t = C \<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R"
    by (cases "match l (get_redex C s)") (auto simp: Let_def)
  with rstep [of l r R C \<sigma>] show ?thesis by auto
qed

    
text \<open>Next derivations\<close>
lemma check_derivation:
  assumes "check_derivation R s seq = Some t"
  shows "(s, t) \<in> rstep R ^^ length seq"
  using assms
proof (induct s seq rule: check_derivation.induct[of _ R])
  case (2 s wit u seq)
  from 2 [simplified] have wit: "check_rstep_wit R wit s u"
    and result: "check_derivation R u seq = Some t"
    by (auto split: if_splits)
  from check_rstep_wit[OF wit]
  have "(s, u) \<in> rstep R" .
  moreover have "(u, t) \<in> rstep R ^^ length seq"
    using 2 and wit and result by auto
  ultimately show ?case
    using relpow_Suc_I2 by fastforce
qed auto

text \<open>Finally loops\<close>
lemma check_loop:
  assumes "check_loop R prf"
  shows "\<exists> t C \<sigma>. (t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
proof (cases "prf")
  case (Loop_Proof t seq C \<sigma>i)
  note ok = assms [unfolded Loop_Proof, simplified]
  from ok have len: "length seq \<noteq> 0" by auto
  obtain Ct_sig where check: "check_derivation R t seq = Some Ct_sig"
    using ok by (auto split: option.splits)
  from check_derivation [OF check]
  have "(t, Ct_sig) \<in> rstep R ^^ length seq" .
  with len have steps: "(t, Ct_sig) \<in> (rstep R)^+"
    by (metis relpow_E trancl_power zero_less_Suc)
  from ok check obtain \<sigma> where t: "Ct_sig = C\<langle>t \<cdot> \<sigma>\<rangle>" by auto
  from steps t show ?thesis by blast
qed

lemma check_loop_not_SN:
  assumes "check_loop R prf"
  shows "\<not> SN (rstep R)"
  using check_loop [OF assms] loop_imp_not_SN_rstep by blast
end