theory Loops
  imports
    Trs
    "$AFP/Abstract-Rewriting/Abstract_Rewriting" (* defines SN *)
begin

text \<open>
  In order to import libraries from the archive of formal proofs (AFP),
  edit @{file "~/.isabelle/Isabelle2016-1/etc/settings"} in
  directory @{dir "~/.isabelle/Isabelle2016-1/etc/"} and add the line:
  init_component "/the/path/to/the/afp"
\<close>

section \<open>What is strong normalization?\<close>

text \<open>No infinite chain\<close>
lemma "SN R = (\<not> (\<exists>f :: nat \<Rightarrow> _. \<forall>i. (f i, f (Suc i)) \<in> R))"
  by fast

text \<open>A relation on which one can perform induction\<close>
lemma "SN R \<Longrightarrow> (\<And>x. (\<And>y. (x, y) \<in> R \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P z"
  using SN_induct [of R P z] by blast

lemma not_SN_I:
  "(\<And>i. (f i, f (Suc i)) \<in> R) \<Longrightarrow> \<not> SN R"
  by (auto simp: SN_def)

section \<open>Loops\<close>

lemma "(t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+ \<Longrightarrow> \<not> (SN ((rstep R)^+))"
proof (rule not_SN_I)
  text \<open>Problem: how to define f in @{thm not_SN_I}?\<close>
oops

text \<open>For loops, we are interested in the following closure property.\<close>
lemma ctxt_subst_closed_rsteps:
  assumes steps: "(s, t) \<in> (rstep R)^+"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+"
  oops

lemma ctxt_subst_closed_rstep:
  assumes step: "(s, t) \<in> rstep R"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> rstep R"
  by (rule ctxt_closed_rstep[OF subst_closed_rstep[OF step]])

text \<open>We can easily generalize and then instantiate later on.\<close>

lemma rtrancl_closed:
  assumes closed: "\<And>a b. (a, b) \<in> R \<Longrightarrow> (f a, f b) \<in> S"
  shows "(a, b) \<in> R^+ \<Longrightarrow> (f a, f b) \<in> S^+"
proof (induct rule: trancl_induct)
  case (base b)
  from closed[OF base] show ?case by simp
next
  case (step b c)
  from closed[OF \<open>(b,c) \<in> R\<close>] step show ?case by auto
qed

lemma ctxt_subst_closed_rsteps:
  assumes steps: "(s, t) \<in> (rstep R)^+"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+"
  by (rule rtrancl_closed[OF ctxt_subst_closed_rstep steps])


context -- \<open>locally fixes terms and assumes properties\<close>
  fixes t :: "('f, 'v) term"
    and C :: "('f, 'v) ctxt"
    and \<sigma> :: "('f, 'v) subst"
    and R :: "('f, 'v) trs"
  assumes loop: "(t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
begin

text \<open>Aim: 
  \<bullet> Write a recursive function @{term "loop_term :: nat \<Rightarrow> ('f, 'v) term"}
    which precisely defines the term after traversing the loop n-times.    

  \<bullet> Prove @{term "(loop_term i, loop_term (Suc i)) \<in> (rstep R)^+"}.
\<close>

fun loop_term :: "nat \<Rightarrow> ('f, 'v) term"
  where
    "loop_term 0 = t" |
    "loop_term (Suc i) = C\<langle>loop_term i \<cdot> \<sigma>\<rangle>"

lemma loop_term_rsteps: "(loop_term i, loop_term (Suc i)) \<in> (rstep R)^+"
proof (induct i)
  case 0
  from loop show ?case by simp
next
  case (Suc i)
  from ctxt_subst_closed_rsteps [OF this]
  show ?case by simp
qed

lemma loop_imp_not_SN_rstep: "\<not> SN (rstep R)"
  using not_SN_I loop_term_rsteps SN_trancl_SN_conv by blast

end

thm loop_imp_not_SN_rstep

end