theory Loops_Demo
  imports
    Trs
    "$AFP/Abstract-Rewriting/Abstract_Rewriting" (* defines SN *)
begin

text \<open>
  In order to import libraries from the archive of formal proofs (AFP),
  edit @{file "~/.isabelle/Isabelle2016-1/etc/settings"} in
  directory @{dir "~/.isabelle/Isabelle2016-1/etc/"} and add the line:
  init_component "/the/path/to/the/afp"
\<close>

section \<open>What is strong normalization?\<close>

text \<open>No infinite chain\<close>
lemma "SN R = (\<not> (\<exists>f :: nat \<Rightarrow> _. \<forall>i. (f i, f (Suc i)) \<in> R))"
  by fast

text \<open>A relation on which one can perform induction\<close>
lemma "SN R \<Longrightarrow> (\<And>x. (\<And>y. (x, y) \<in> R \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P z"
  using SN_induct [of R P z] by blast

lemma not_SN_I:
  "(\<And>i. (f i, f (Suc i)) \<in> R) \<Longrightarrow> \<not> SN R"
  by (auto simp: SN_def)

section \<open>Loops\<close>

context -- \<open>locally fixes terms and assumes properties\<close>
  fixes t :: "('f, 'v) term"
    and C :: "('f, 'v) ctxt"
    and \<sigma> :: "('f, 'v) subst"
    and R :: "('f, 'v) trs"
  assumes loop: "(t, C\<langle> t \<cdot> \<sigma> \<rangle>) \<in> (rstep R)^+"
begin

lemma "\<not> SN ((rstep R)^+)" 
proof (rule not_SN_I)
  oops
  
fun loop_term :: "nat \<Rightarrow> ('f, 'v) term"
  where
    "loop_term 0 = undefined" |
    "loop_term (Suc i) = undefined"

lemma loop_term_rsteps: "(loop_term i, loop_term (Suc i)) \<in> (rstep R)^+"
  oops
    
lemma loop_imp_not_SN_rstep: "\<not> SN (rstep R)"
  oops

end
end