theory Loops_Ex (* exercise version of Loops.thy *)
  imports
    Trs
    "$AFP/Abstract-Rewriting/Abstract_Rewriting" (* defines SN *)
begin

text \<open>
  In order to import libraries from the archive of formal proofs (AFP),
  edit @{file "~/.isabelle/Isabelle2016-1/etc/settings"} in
  directory @{dir "~/.isabelle/Isabelle2016-1/etc/"} and add the line:
  init_component "/the/path/to/the/afp"
\<close>

section \<open>What is strong normalization?\<close>

text \<open>No infinite chain\<close>
lemma "SN R = (\<not> (\<exists>f :: nat \<Rightarrow> _. \<forall>i. (f i, f (Suc i)) \<in> R))"
  by fast

text \<open>A relation on which one can perform induction\<close>
lemma "SN R \<Longrightarrow> (\<And>x. (\<And>y. (x, y) \<in> R \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P z"
  using SN_induct [of R P z] by blast

lemma not_SN_I:
  "(\<And>i. (f i, f (Suc i)) \<in> R) \<Longrightarrow> \<not> SN R"
  by (auto simp: SN_def)

section \<open>Loops\<close>

definition loop :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) subst \<Rightarrow> bool"
  where
    "loop R t C \<sigma> = ((t, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+)"

lemma "loop R t C \<sigma> \<Longrightarrow> \<not> (SN ((rstep R)^+))"
proof (rule not_SN_I)
  text \<open>Problem: how to define f in @{thm not_SN_I}?\<close>
oops

text \<open>For loops, we are interested in the following closure property.\<close>
lemma ctxt_subst_closed_rsteps:
  assumes steps: "(s, t) \<in> (rstep R)^+"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+"
  oops

lemma ctxt_subst_closed_rstep:
  assumes step: "(s, t) \<in> rstep R" 
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> rstep R"
proof - -- \<open>Start proof with - to just state arbitrary intermediate facts
  as in induction proof. The final goal is "?thesis" and must be shown
  via "show"\<close>
  from subst_closed_rstep[OF step]
  have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R" .
  from ctxt_closed_rstep[OF this] show ?thesis .
qed


text \<open>We can easily generalize and then instantiate later on.\<close>

lemma rtrancl_closed:
  assumes closed: "\<And>a b. (a, b) \<in> R \<Longrightarrow> (f a, f b) \<in> S"
  shows "(a, b) \<in> R^+ \<Longrightarrow> (f a, f b) \<in> S^+"
  find_theorems "_^+" name: induct -- \<open>find_theorems for searching (induction rule),
   application "find_theorems term1 term2 ..." where term_i may contain wildcard _\<close>
proof (induct rule: trancl_induct)
  case (base b)
  from closed[OF base] show ?case by simp
next
  case (step b c)
  from closed[OF \<open>(b,c) \<in> R\<close>] step show ?case by auto
qed

lemma ctxt_subst_closed_rsteps:
  assumes steps: "(s, t) \<in> (rstep R)^+"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+"
  by (rule rtrancl_closed[OF ctxt_subst_closed_rstep steps])


context -- \<open>locally fixes terms and assumes properties\<close>
  fixes t :: "('f, 'v) term"
    and C :: "('f, 'v) ctxt"
    and \<sigma> :: "('f, 'v) subst"
    and R :: "('f, 'v) trs"
  assumes loop: "loop R t C \<sigma>"
begin

subsection \<open>Exercise\<close>

text \<open>
  prove that R is non-terminating!

  \<bullet> Write a recursive function @{term "loop_term :: nat \<Rightarrow> ('f,'v)term"}
    which precisely defines the term after traversing the loop n-times.
    To this end, use pattern matching on @{const Zero} and @{const Suc}.
    The function definition can use @{term t}, @{term C}, ... from the
    context.

  \<bullet> Prove @{term "(loop_term i, loop_term (Suc i)) \<in> (rstep R)^+"}.
    Hints: The defining equation of @{const loop} is @{thm loop_def}.
    One can locally add a theorem thm to the simplifier, e.g., as in
    "note [simp] = thm" or "by (simp add: thm)"

  \<bullet> Use @{command find_theorems} to find theorems which combine
    strong normalization with the transitive closure.
\<close>

fun loop_term :: "nat \<Rightarrow> ('f, 'v) term"
  where
    "loop_term 0 = undefined" |
    "loop_term (Suc i) = undefined"

lemma loop_term_rsteps: "(loop_term i, loop_term (Suc i)) \<in> (rstep R)^+"
  sorry

lemma loop_imp_not_SN_rstep: "\<not> SN (rstep R)"
  sorry

end

thm loop_imp_not_SN_rstep

end