theory Matching_Completeness
  imports Matching
begin

section \<open>Completeness of our Matching Algorithm\<close>

definition "matchers E = {\<sigma>. \<forall>(p, t) \<in> E. p \<cdot> \<sigma> = t}"

lemma matchers_empty [simp]: "matchers {} = UNIV"
  by (simp add: matchers_def)

lemma matchers_insert [simp]:
  "matchers (insert (s, t) P) = {\<sigma>. s \<cdot> \<sigma> = t} \<inter> matchers P"
  by (auto simp: matchers_def)

lemma matchers_zip [simp]:
  assumes "length ss = length ts"
  shows "matchers (set (zip ss ts)) = {\<sigma>. map (subst_term \<sigma>) ss = ts}"
  using assms by (induct ss ts rule: list_induct2) auto

definition "mos V \<sigma> x = (if x \<in> V then Some (\<sigma> x) else None)"

lemma mos_empty [simp]: "mos {} \<sigma> = Map.empty" by (auto simp: mos_def)

lemma mos_singleton [simp]: "mos {x} \<sigma> = [x \<mapsto> \<sigma> x]" by (auto simp: mos_def)

lemma dom_mos [simp]: "dom (mos V \<sigma>) = V" by (auto simp: mos_def split: if_splits)

lemma mos_union [simp]: "mos (V \<union> W) \<sigma> = mos V \<sigma> ++ mos W \<sigma>"
  by (auto simp: mos_def map_add_def)

lemma match_instance:
  assumes "\<forall>x \<in> dom \<sigma>. \<sigma> x = Some (\<tau> x)"
  shows "match_term \<sigma> t (t \<cdot> \<tau>) = Some (\<sigma> ++ mos (vars_term t) \<tau>)"
    and "match_list \<sigma> ts (map (subst_term \<tau>) ts) = Some (\<sigma> ++ mos (\<Union>set (map vars_term ts)) \<tau>)"
  using assms
proof (induct \<sigma> t "t \<cdot> \<tau>" and \<sigma> ts "map (subst_term \<tau>) ts" rule: match_term_match_list.induct)
  case (5 \<sigma> t ts s ss)
  let ?\<sigma> = "\<sigma> ++ mos (vars_term t) \<tau>"
  have "match_list ?\<sigma> ts (map (subst_term \<tau>) ts) = Some (?\<sigma> ++ mos (\<Union>set (map vars_term ts)) \<tau>)"
    using 5 by (intro 5) (simp, auto simp: mos_def map_add_Some_iff)
  with 5 show ?case by simp
qed auto

lemma match_complete:
  shows "match_term \<sigma> s t = None \<Longrightarrow> \<forall>x \<in> dom \<sigma>. \<forall>\<tau> \<in> matchers {(s, t)}. \<sigma> x = Some (\<tau> x) \<Longrightarrow>
    matchers {(s, t)} = {}"
    and "match_list \<sigma> ss ts = None \<Longrightarrow> \<forall>x \<in> dom \<sigma>. \<forall>\<tau> \<in> matchers (set (zip ss ts)). \<sigma> x = Some (\<tau> x) \<Longrightarrow>
    length ss = length ts \<Longrightarrow> matchers (set (zip ss ts)) = {}"
   apply (induct \<sigma> s t and \<sigma> ss ts rule: match_term_match_list.induct)
        apply (auto split: if_splits)
  apply (auto split: option.splits simp: match_instance)
  by (smt Int_Collect domIff map_add_None map_add_Some_iff match_instance mem_Collect_eq mos_def option.discI)

end
