theory Term_Demo (* demo version of Term.thy *)
  imports Main
begin

section \<open>Definition of Types\<close>

subsection \<open>@{command datatype}\<close>

text \<open>
  \<bullet> defines new algebraic datatype as in functional programming

  \<bullet> several automatically generated functions:
    selectors, map, set-functions, etc.

  \<bullet> provides induction schema, simplification rules, etc.

  \<bullet> @{doc "datatypes"}
\<close>


text \<open>
  Definition of first-order terms for
  function symbols of type @{typ 'f}
  and variables of type @{typ 'v}
\<close>

datatype ('f, 'v) "term" = TODO

text \<open>
  Invoke @{command print_theorems} directly
  after a declaration to see new theorem that have been derived
\<close>


text \<open>Induction rule\<close>


text \<open>Case analysis\<close>

text \<open>
  @{term vars_term} is set-function which collects all
  variables in a term
\<close>


text \<open>Simplification rules\<close>



subsection \<open>@{command type_synonym}\<close>

text \<open>
  \<bullet> for defining type abbreviations
  \<bullet> just syntactic sugar, no new type is created
  \<bullet> no new theorems will be created
\<close>

text \<open>Substitutions are functions from variables to terms\<close>

type_synonym ('f, 'v) subst = unit

subsection \<open>@{command typedef}\<close>

text \<open>
  \<bullet> for defining subtypes
  \<bullet> internally used by @{command datatype}
  \<bullet> not further relevant in this course
\<close>

typedef 'f ground_term = "UNIV :: 'f set" by auto


section \<open>Definition of Functions\<close>

text \<open>
  Five different definition principles

  \<bullet> @{command abbreviation}
     \<bullet> similar to type-synonym
     \<bullet> just syntactic sugar

  \<bullet> @{command definition}
     \<bullet> non-recursive
     \<bullet> no pattern matching
     \<bullet> equality not added to simplification rules

  \<bullet> @{command function}, @{command fun}
     \<bullet> recursive
     \<bullet> manual or automatic termination proof
     \<bullet> pattern matching
     \<bullet> induction scheme
     \<bullet> @{doc "functions"}

  \<bullet> @{command primrec}
     \<bullet> primitive recursion
     \<bullet> faster than @{command fun}
     \<bullet> @{doc "datatypes"}

  \<bullet> @{command "partial_function"}
     \<bullet> permits nontermination
     \<bullet> no pattern matching
     \<bullet> syntactic restrictions (tailrec, option)
     \<bullet> @{doc "isar-ref"}, Chapter 11.2.2
\<close>

definition subst_term :: "('f, 'v) subst \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
  where
    "subst_term \<sigma> x = undefined"

print_theorems

abbreviation apply_subst_term (infixl "\<cdot>" 67)
  where
    "t \<cdot> \<sigma> \<equiv> subst_term \<sigma> t"

end
