theory Trs_Demo (* demo version of Trs.thy *)
  imports Term
begin

section \<open>First-Order Term Rewriting\<close>

text \<open>
  \<bullet> @{command inductive} and @{command inductive_set} for inductive
    predicates and inductively defined sets

  \<bullet> provides induction-rules, introduction-rules, etc.
\<close>

inductive_set rstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs" for R
  where
    rstep: "(l,r) \<in> R \<Longrightarrow> (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R"

text \<open>
  proofs by induction @{method induct} or @{method induction}}

  \<bullet> take elements on which to perform induction

  \<bullet> specify which elements should be generalized: "arbitrary: x"

  \<bullet> take optional induction rule, e.g., from
     \<bullet> @{command datatype}
     \<bullet> @{command inductive} or @{command inductive_set}
     \<bullet> @{command fun} or @{command function}
     \<bullet> manually proven induction rule
\<close>

lemma example_induction: "vars_term s = {} \<Longrightarrow> s \<cdot> \<sigma> = s \<cdot> \<delta>"
  by (induct s, auto)

lemma example_induction': "vars_term s = {} \<Longrightarrow> s \<cdot> \<sigma> = s \<cdot> \<delta>"
proof (induct s arbitrary: \<delta> rule: term.induct)
  case (Var x \<delta>)  
  from Var have x_empty: "{x} = {}" by simp
  from x_empty have False by simp
  then show ?case ..
next
  case (Fun f ts \<delta>)
  thm Fun
  thm Fun.hyps
  thm Fun.prems
  { 
    fix t
    assume t: "t \<in> set ts" 
    then have "vars_term t = {}" using Fun.prems by auto
    from Fun.hyps[OF t this] have "t \<cdot> \<sigma> = t \<cdot> \<delta>" .
  } 
  from this show ?case by simp 
qed

text \<open>Back to our goal\<close>

lemma ctxt_closed_rstep: "(s, t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep R"
  sorry

end