theory Trs_Ex (* exercise version of Trs.thy *)
  imports Term
begin

section \<open>First-Order Term Rewriting\<close>

text \<open>
  \<bullet> @{command inductive} and @{command inductive_set} for inductive
    predicates and inductively defined sets

  \<bullet> provides induction-rules, introduction-rules, etc.
\<close>

inductive_set rstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term rel" for R
  where
    rstep: "(l, r) \<in> R \<Longrightarrow> (C \<langle> l \<cdot> \<sigma> \<rangle>, C \<langle> r \<cdot> \<sigma> \<rangle>) \<in> rstep R"

print_theorems

text \<open>
  proofs by induction @{method induct} or @{method induction}}

  \<bullet> take elements on which to perform induction

  \<bullet> specify which elements should be generalized: "arbitrary: x"

  \<bullet> take optional induction rule, e.g., from
     \<bullet> @{command datatype}
     \<bullet> @{command inductive} or @{command inductive_set}
     \<bullet> @{command fun} or @{command function}
     \<bullet> manually proven induction rule
\<close>

lemma example_induction: "vars_term s = {} \<Longrightarrow> s \<cdot> \<sigma> = s \<cdot> \<delta>"
  by (induct s) (auto)

lemma example_induction': "vars_term s = {} \<Longrightarrow> s \<cdot> \<sigma> = s \<cdot> \<delta>"
proof (induct s arbitrary: \<delta> rule: term.induct)
  print_cases -- \<open>figure out cases\<close>
  case (Var x \<delta>) -- \<open>each case is started by "case",
    choose own parameter names, e.g., here the variable name and the arbitrary \<delta>\<close>
  from Var have x_empty: "{x} = {}" by simp
  from x_empty have False by simp
  then show ?case ..
next
  case (Fun f ts \<delta>)
  thm Fun
  thm Fun.hyps
  thm Fun.prems
  { 
    fix t
    assume t: "t \<in> set ts" 
    then have "vars_term t = {}" using Fun.prems by auto
    from Fun.hyps[OF t this] have "t \<cdot> \<sigma> = t \<cdot> \<delta>" .
  } 
  from this show ?case by simp 
qed

text \<open>Back to our goal\<close>

lemma ctxt_closed_rstep: "(s, t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep R"
proof (induct s t rule: rstep.induct)
  case (rstep l r D \<sigma>)
  show ?case using rstep.rstep[OF rstep, of "C \<circ>\<^sub>c D" \<sigma>] (* not by simp *)
oops -- \<open>aborts proof attempt\<close>

lemma ctxt_ctxt_compose [simp]: -- \<open>[simp]: use lemma in simplifier\<close>
  "(C \<circ>\<^sub>c D)\<langle>t\<rangle> = C\<langle>D\<langle>t\<rangle>\<rangle>"
  by (induct C) (auto)

lemma ctxt_closed_rstep: "(s, t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> rstep R"
proof (induct s t rule: rstep.induct)
  case (rstep l r D \<sigma>)
  show ?case using rstep.rstep[OF rstep, of "C \<circ>\<^sub>c D" \<sigma>] by simp
qed

subsection \<open>Exercise: closure under substitutions\<close>

fun
  subst_ctxt :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) ctxt" (infixl "\<cdot>\<^sub>c" 67)
  where
    "\<box> \<cdot>\<^sub>c \<sigma> = \<box>"
  | "(More f ss1 D ss2) \<cdot>\<^sub>c \<sigma> = More f (map (\<lambda>t. t \<cdot> \<sigma>) ss1) (D \<cdot>\<^sub>c \<sigma>) (map (\<lambda>t. t \<cdot> \<sigma>) ss2)"

definition
  subst_compose :: "('f, 'v) subst \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
  (infixl "\<circ>\<^sub>s" 75)
  where
    "\<sigma> \<circ>\<^sub>s \<tau> = (\<lambda>x. (\<sigma> x) \<cdot> \<tau>)"

thm subst_compose_def

text \<open>
  \<bullet> use induction on @{const rstep}, similar to @{thm ctxt_closed_rstep}

  \<bullet> which witness is required when instantiating @{thm rstep.rstep}?

  \<bullet> if proof is not automatic after providing the right witness,
    first prove some equalities and add them to the simplifier
\<close>

lemma subst_closed_rstep: "(s, t) \<in> rstep R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R"
  sorry

end